/***************************************************************************
                    mymap_canvas.cpp  -  displays mindmaps
                             -------------------                                         
    begin                : Tue Nov 23 1999                                           
    copyright            : (C) 1999 by Jost Schenck                         
    email                : jost@schenck.de                                     
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the ARTISTIC LICENSE.                           * 
 *                                                                         *
 ***************************************************************************/

#include <qaccel.h>
#include <qpainter.h>
#include <qpixmap.h>
#include <qpopupmenu.h>
#include <qlineedit.h>
#include <qstringlist.h>
#include <kaction.h>
#include <klocale.h>

#include "mymap_map.h"
#include "mymap_doc.h"
#include "mymap_canvas.h"
#include "mymap_graphnode.h"
#include "mymap_item.h"
#include "mymap_selection.h"
#include "propertiesdlg.h"

MyMapCanvas::MyMapCanvas(QWidget *parent, const char *name ) : QWidget(parent,name)
{
	pMap = 0;
	pSelection = new MyMapSelection();
	pCurrent = 0;
	bEditMode = false;
	pEditNode = 0;

	bDragging = false; bMoved = false; bResizing = false;

	pEditor = new QLineEdit( this );
	pEditor->hide();
	pEditor->setFrame( false );
	connect( pEditor, SIGNAL( textChanged( const QString& ) ), this, SLOT( editorTextChanged( const QString& ) ) );
	
	initActions();
	
	// Initialize context menus
	pNodePopup = new QPopupMenu( this );
		acNewChild->plug( pNodePopup );
		acDeleteSelection->plug( pNodePopup );
		acNodeStyle->plug( pNodePopup );
		acEditCurrentProperties->plug( pNodePopup );
	
	// Initialize accel keys
	QAccel* acc = new QAccel( this );
	acc->insertItem( Key_Return, 200 );
	acc->insertItem( Key_Escape, 201 );
	acc->insertItem( Key_Left, 202 );
	acc->insertItem( Key_Right, 203 );
	acc->insertItem( Key_Up, 204 );
	acc->insertItem( Key_Down, 205 );
	acc->insertItem( Key_Return+CTRL, 206);
	acc->insertItem( Key_Return+ALT, 207);
	acc->insertItem( Key_Delete, 208 );
	connect(acc, SIGNAL(activated( int )), this, SLOT(slotAccel( int )) );
}

MyMapCanvas::~MyMapCanvas()
{
}

void MyMapCanvas::initActions()
{
	acNewChild = new KAction( i18n("create &new child"), 0, this,
	                         SLOT( appendNewChild() ), this, "newchild");
	acDeleteSelection = new KAction( i18n("&delete"), 0, this,
	                         SLOT( deleteSelection() ), this, "delete");
	acNodeStyle = new KSelectAction( i18n("node &style"), 0, this, "selectstyle" );
	connect( acNodeStyle, SIGNAL( activated( const QString& ) ), this, SLOT(changeNodeStyle(const QString&)) );
	acNodeStyle->setEditable( false );
	QStringList items = QStringList() << i18n("Ellipse") << i18n("Fork") << i18n("Rectangle") << i18n("Rounded Rectangle") << i18n("Transparent");
	acNodeStyle->setItems( items );
	acEditCurrentProperties = new KAction( i18n("&edit properties"), 0, this,
	                         SLOT( editCurrentProperties() ), this, "editprops");
}

void MyMapCanvas::setMap( MyMapMap* m )
{
	debug("MyMapCanvas::setMap ");
	// FIXME! Canvas should only be connected to View, View to Canvas and Doc
	if ( pMap )
	{
		disconnect( pMap, SIGNAL( update() ), this, SLOT( updateView() ) );
		disconnect( this, SIGNAL(change()), pMap->document(), SLOT( docChanged() ) );
		disconnect( pMap->document(), SIGNAL(updateSignal()), this, SLOT( updateView() ) );
	}
	debug("MyMapCanvas::setMap 1");
	pMap = m;
	if ( pMap )
	{
		connect( pMap, SIGNAL(update()), this, SLOT(updateView() ) );
		connect( this, SIGNAL(change()), pMap->document(), SLOT( docChanged() ) );
		connect( pMap->document(), SIGNAL(updateSignal()), this, SLOT( updateView() ) );
	}
	debug("MyMapCanvas::setMap 2");
	if ( pSelection )
		pSelection->clear();
	debug("MyMapCanvas::setMap 3");
	pCurrent = 0;	
	if ( pMap )
		mapPropertiesChanged();
	debug("MyMapCanvas::setMap 4");
	repaint();
	debug("MyMapCanvas::setMap ENDE");
}

void MyMapCanvas::paintEvent( QPaintEvent* )
{
//	debug("paintEvent ---------------------");
	if ( pMap && pMap->rootNode() )
	{
//		debug("1");
		QPainter* p = new QPainter( this );
		pMap->rootNode()->paint( p, false );
//		debug("2");
		pSelection->paint( p );
		if ( pCurrent )
		{
			QPen pen;
			pen.setStyle( Qt::DotLine );
			pen.setColor( Qt::black );
			pen.setWidth( 1 );
			p->setPen( pen );
			p->setBrush( Qt::NoBrush );
			p->drawRect( pCurrent->x(), pCurrent->y(), pCurrent->width(), pCurrent->height());			
		}
		delete p;
	}
//	debug("paintEvent ---------------------");
}

void MyMapCanvas::mousePressEvent( QMouseEvent* e )
{
	/* TODO */
	if ( pMap )
	{
		editMode( false );
		rh = pSelection->getHandle( e->pos() );
		if ( rh != rhNone )
		{
			debug("RESIZE HANDLE: %i", rh );
			if ( e->button() & Qt::LeftButton )
			{
				bResizing = true;
				bMoved = false;
			}
		}
		else
		{
			// FIXME: get rid of dynamic_cast
			MyMapGraphNode* node = dynamic_cast<MyMapGraphNode*>(pMap->itemAt( e->pos()));
			if ( node )
			{
				if ( !(e->state() & Qt::ControlButton) && !pSelection->isSelected( node ))
				{
					pSelection->clear();
					pSelection->addObject( node );
				}
				else
					pSelection->addObject( node );
				setCurrent( node );		
				if ( e->button() & Qt::RightButton )
					pNodePopup->popup( this->mapToGlobal( e->pos() ) );
				else
				if ( e->button() & Qt::LeftButton )
				{
					bDragging = true;
					bMoved = false;
					xDragOffset = e->x() - pSelection->x();
					yDragOffset = e->y() - pSelection->y();
				}
			}
			else
			{
				pSelection->clear();
			}
		}
		repaint();
	}
}

void MyMapCanvas::mouseReleaseEvent( QMouseEvent* e )
{
	if ( bDragging )
	{
		// FIXME: get rid of dynamic_cast
		MyMapGraphNode* node = dynamic_cast<MyMapGraphNode*>(pMap->itemAt( e->pos() ));
		if ( bMoved )
		{
			if ( node && !pSelection->isSelected(node) )
			{
				pSelection->reparent( node );
				pSelection->setGeometry();
			}
			else
				pSelection->move( e->x() - xDragOffset, e->y() - yDragOffset );
			emit change();
		}
//		else
//			if ( node ) pSelection->toggleObject( node );
	}
	else if ( bResizing && bMoved )
	{
		debug("RESIZE RELEASED --------------");
		pSelection->resize( rh, e->pos() );	
	}
	
	bDragging = false;
	bResizing = false;
	bMoved = false;
}

void MyMapCanvas::mouseDoubleClickEvent( QMouseEvent* e )
{
	if ( pMap )
	{
		// FIXME: get rid of dynamic_cast
//		MyMapGraphNode* node = dynamic_cast<MyMapGraphNode*>(pMap->itemAt( e->pos() ));
		MyMapItem* node = pMap->itemAt( e->pos() );
		if ( node )
			editMode();
		else
			appendNewChildAt( e->x(), e->y() );
	}
}

void MyMapCanvas::mouseMoveEvent( QMouseEvent* )
{
	bMoved = true;
}

MyMapGraphNode* MyMapCanvas::appendNewChildAt( QCOORD x_, QCOORD y_ )
{
	MyMapGraphNode* node = 0;
	if ( pCurrent )
	{
		node = new MyMapGraphNode( pCurrent );
		debug("appendNewChildAt 1");
		node->setStyle( pCurrent->standardChildNodeStyle() );
		debug("appendNewChildAt 2");
		if ( pCurrent->style() != mmFORKSTYLE )
			node->moveTo( x_, y_ );
		debug("appendNewChildAt 3");
		emit change();
		debug("appendNewChildAt 4");
	}
	return node;
}

void MyMapCanvas::appendNewChild()
{
	// neues Child an current
	if ( pCurrent )
		appendNewChildAt( pCurrent->x() + 30, pCurrent->y() + 30 ); // better positioning
}

void MyMapCanvas::deleteSelection()
{
	QListIterator<MyMapItem> it ( pSelection->selection() );
	for (; it.current(); ++it)
		if ( it.current()!=pMap->rootNode() ) // change this when it is possible to have none or multiple roots
			delete ( it.current() );
	pSelection->clear();
	setCurrent(0);
	emit change();
}

void MyMapCanvas::changeNodeStyle( const QString& text )
{
	if ( pCurrent ) // HACK, see setCurrent()
	{
		QListIterator<MyMapItem> it ( pSelection->selection() );
		KMMNODESTYLE sty = mmELLIPSE;
		if ( text == i18n("Ellipse") )
			sty = mmELLIPSE;
		if ( text == i18n("Fork") )
			sty = mmFORKSTYLE;
		if ( text == i18n("Rectangle") )
			sty = mmRECTANGLE;
		if ( text == i18n("Rounded Rectangle") )
			sty = mmROUNDRECT;
		if ( text == i18n("Transparent") )
			sty = mmTRANSPARENT;
		for (; it.current(); ++it)
			if ( it.current()->rtti() == GraphNode )
				dynamic_cast<MyMapGraphNode*>(it.current())->setStyle( sty );
		emit change();
	}
}

void MyMapCanvas::editMode( bool b )
{
	bEditMode = b;
	if ( bEditMode )
	{
		pSelection->clear();
		setEditorGeometry();
		// TODO: save original Text in sOriginalText
		pEditor->setText( pCurrent->text() );
		// FIXME: this doesn't work
		pEditor->setPalette( QPalette( pCurrent->backColor(), pCurrent->backColor() ) );
		pEditor->show();
		pEditor->setFocus();
	}
	else
	{
		pEditor->hide();
		emit checkTitles();
	}
}

void MyMapCanvas::updateView() { debug("UPDATE VIEW"); repaint(); }

void MyMapCanvas::newDocLoaded()
{
	debug("MyMapCanvas::newDocLoaded!");
	setCurrent( pMap->rootNode() );
	pSelection->clear();
	debug("MyMapCanvas::newDocLoaded ENDE");
}

void MyMapCanvas::editorTextChanged( const QString& text )
{
	pCurrent->setText( text );
	setEditorGeometry();
	emit change();
}

void MyMapCanvas::editCurrentProperties()
{
	// FIXME: editing of the whole selection
	PropertiesDlg* dlg = new PropertiesDlg( 0, pCurrent );
	dlg->showTab( 0 );
	dlg->show();
}

void MyMapCanvas::editCurrentConProperties()
{
	// FIXME: editing of the whole selection
	PropertiesDlg* dlg = new PropertiesDlg( 0, pCurrent );
	dlg->showTab( 1 );
	dlg->show();
}

void MyMapCanvas::slotAccel( int id )
{
  MyMapItem* pTmpItem = 0;
  MyMapGraphNode* pTmpNode = 0;
  switch ( id )
  {
    case 200:    // Enter pressed
      if ( bEditMode )
        editMode( false );
      else if ( pCurrent )
        editMode();
      break;
    case 201:    // Escape pressed
//        emit changeItemText( pEditNode->getDataObject(), sEditOrigText );
			editMode( false );
      break;
    case 202:    // Left pressed
      if ( pCurrent )
      {
				if ( bEditMode )
					pEditor->setCursorPosition( pEditor->cursorPosition() - 1 );
				else
				if ( pCurrent->isRightSide() == 1 && pCurrent->parent() )
					setCurrent( pCurrent->parent() );
				else if ( pCurrent->children().getFirst() && pCurrent->children().getFirst()->rtti() == GraphNode )
					setCurrent( dynamic_cast<MyMapGraphNode*>(pCurrent->children().getFirst()) );
			}
      repaint();
      break;
    case 203:    // Right pressed
      if ( pCurrent )
      {
				if ( bEditMode )
					pEditor->setCursorPosition( pEditor->cursorPosition() + 1 );
				else
				if ( pCurrent->isRightSide() == -1 && pCurrent->parent() )
					setCurrent( pCurrent->parent() );
				else if ( pCurrent->children().getFirst() && pCurrent->children().getFirst()->rtti() == GraphNode )
					setCurrent( dynamic_cast<MyMapGraphNode*>(pCurrent->children().getFirst()) );
			}
			repaint();
      break;
		case 204:    // Up
			// TODO: change this by adding MyMapGraphNode::nextSibling( MyMapGraphNode*) and prevSibling(...)
			if ( pCurrent && pCurrent->parent() )
			{
				pTmpItem = pCurrent->parent()->childAt( pCurrent->parent()->childPos( pCurrent ) - 1);
				if ( pTmpItem && pTmpItem->rtti() == GraphNode )
					// TODO: get rid of this and the following casts when pCurrent is MyMapItem
					setCurrent( dynamic_cast<MyMapGraphNode*>(pTmpItem) );
			}
			repaint();
			break;
		case 205:    // Down
			if ( pCurrent && pCurrent->parent() )
			{
				pTmpItem = pCurrent->parent()->childAt( pCurrent->parent()->childPos( pCurrent ) + 1);
				if ( pTmpItem && pTmpItem->rtti() == GraphNode )
					setCurrent( dynamic_cast<MyMapGraphNode*>(pTmpItem) );
			}
			repaint();
			break;
		case 206:    // CTRL+Return
			if ( pCurrent )
			{
				pTmpNode = pCurrent;
				setCurrent( new MyMapGraphNode( pCurrent ) );
				pCurrent->setStyle( pTmpNode->standardChildNodeStyle() );
				if ( !pTmpNode->isAlignChildren() )
					pCurrent->moveTo( pTmpNode->x() + ( (pTmpNode->isRightSide()==1)?(pTmpNode->width()+30):(-30 - pCurrent->width())),
					                  pTmpNode->y() + 20 );
				emit change();
				repaint();
				editMode();
			}
			break;
		case 207:		// ALT+Return
			if ( pCurrent && pCurrent->parent() )
			{
				// FIXME: insert child right after current item
				pTmpNode = pCurrent;
				setCurrent( new MyMapGraphNode( pCurrent->parent() ) );
				if ( !pCurrent->parent()->isAlignChildren() )
					pCurrent->moveTo( pCurrent->parent()->x() + ( (pCurrent->parent()->isRightSide()==1)?(pCurrent->parent()->width()+30):(-30 - pCurrent->width())),
					                  pTmpNode->y() + pTmpNode->height() + 15 );
				emit change();
				repaint();
				editMode();
			}
			break;
		case 208:		// Delete
			if ( !bEditMode )
			{
				debug("0");
				if ( pSelection->isEmpty() && pCurrent && pCurrent->parent() )
				{
					debug("1");
					pTmpNode = pCurrent->parent();
					pTmpNode->removeChild( pCurrent );
					debug("2");
					delete pCurrent;
					pCurrent = pTmpNode;
					debug("3");
					emit change();
					debug("4");
				}
				else
				{
					deleteSelection();
					pCurrent = 0;
					// what should pCurrent be now?
				}
				repaint();				
			}
			else
				pEditor->del();
			break;
	}
}

void MyMapCanvas::setCurrent( MyMapGraphNode* c )
{
	pCurrent = 0; // Do this so the change of the selectAction doesn't change any nodes
	if ( c )
	{
		// make this nicer
		if      ( c->style() == mmELLIPSE)     acNodeStyle->setCurrentItem( 0 );
		else if ( c->style() == mmFORKSTYLE)   acNodeStyle->setCurrentItem( 1 );
		else if ( c->style() == mmRECTANGLE)   acNodeStyle->setCurrentItem( 2 );
		else if ( c->style() == mmROUNDRECT)   acNodeStyle->setCurrentItem( 3 );
		else if ( c->style() == mmTRANSPARENT) acNodeStyle->setCurrentItem( 4 );
	}
	pCurrent = c;
	if ( bEditMode )
	{
		editMode( true );
	}
}

void MyMapCanvas::setEditorGeometry()
{
	pEditor->setGeometry( pCurrent->editorBounds() );
}

void MyMapCanvas::mapPropertiesChanged()
{
	setBackgroundMode( FixedColor );
	setBackgroundColor( pMap->bgColor() );
}




